@@ -19,10 +19,10 @@
  *************************************************************************************/
 #ifdef HAVE_GNUTLS
 
+#include <ma_global.h>
 #include <gnutls/gnutls.h>
 #include <gnutls/x509.h>
 #include <gnutls/abstract.h>
-#include <ma_global.h>
 #include <ma_sys.h>
 #include <ma_common.h>
 #include <ma_pvio.h>
@@ -50,12 +50,6 @@ static int my_verify_callback(gnutls_session_t ssl);
 
 char tls_library_version[TLS_VERSION_LENGTH];
 
-struct st_gnutls_data {
-  MYSQL *mysql;
-  gnutls_privkey_t key;
-  gnutls_pcert_st cert;
-};
-
 struct st_cipher_map {
   unsigned char sid[2];
   const char *iana_name;
@@ -799,18 +793,6 @@ const struct st_cipher_map tls_ciphers[]=
     NULL}
 };
 
-/* free data assigned to the connection */
-static void free_gnutls_data(struct st_gnutls_data *data)
-{
-  if (data)
-  {
-    if (data->key)
-      gnutls_privkey_deinit(data->key);
-    gnutls_pcert_deinit(&data->cert);
-    free(data);
-  }
-}
-
 /* map the gnutls cipher suite (defined by key exchange algorithm, cipher
    and mac algorithm) to the corresponding OpenSSL cipher name */
 static const char *openssl_cipher_name(gnutls_kx_algorithm_t kx,
@@ -896,12 +878,7 @@ static void ma_tls_set_error(MYSQL *mysql, void *ssl, int ssl_errno)
   char  ssl_error[MAX_SSL_ERR_LEN];
   const char *ssl_error_reason;
   MARIADB_PVIO *pvio= mysql->net.pvio;
-
-  if (!ssl_errno)
-  {
-    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, "Unknown SSL error");
-    return;
-  }
+  int save_errno= errno;
 
   /* give a more descriptive error message for alerts */
   if (ssl_errno == GNUTLS_E_FATAL_ALERT_RECEIVED)
@@ -912,20 +889,20 @@ static void ma_tls_set_error(MYSQL *mysql, void *ssl, int ssl_errno)
     alert_name= gnutls_alert_get_name(alert_desc);
     snprintf(ssl_error, MAX_SSL_ERR_LEN, "fatal alert received: %s",
              alert_name);
-    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, 0, 
-                   ssl_error);
+    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, 0, ssl_error);
     return;
   }
 
-  if ((ssl_error_reason= gnutls_strerror(ssl_errno)))
+  if (ssl_errno && (ssl_error_reason= gnutls_strerror(ssl_errno)))
   {
-    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, 0, 
+    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, 0,
                    ssl_error_reason);
     return;
   }
-  snprintf(ssl_error, MAX_SSL_ERR_LEN, "SSL errno=%d", ssl_errno);
-  pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, 
-                 ssl_error);
+
+  strerror_r(save_errno, ssl_error, MAX_SSL_ERR_LEN);
+  pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, "TLS/SSL error: %s (%d)",
+                  ssl_error, save_errno);
 }
 
 
@@ -1025,11 +1002,15 @@ static size_t ma_gnutls_get_protocol_version(const char *tls_version_option,
     strcat(tls_versions, ":+VERS-TLS1.1");
   if (strstr(tls_version_option, "TLSv1.2"))
     strcat(tls_versions, ":+VERS-TLS1.2");
+#if GNUTLS_VERSION_NUMBER > 0x030605
+  if (strstr(tls_version_option, "TLSv1.3"))
+    strcat(tls_versions, ":+VERS-TLS1.3");
+#endif
 end:
   if (tls_versions[0])
     snprintf(priority_string, prio_len - 1, "-VERS-TLS-ALL%s:NORMAL", tls_versions);
   else
-    strncpy(priority_string, "NORMAL", prio_len - 1);
+    strncpy(priority_string, "NORMAL:+VERS-ALL", prio_len - 1);
   return strlen(priority_string);
 }
 
@@ -1049,7 +1030,7 @@ static int ma_gnutls_set_ciphers(gnutls_session_t ssl,
 
   token= strtok((char *)cipher_str, ":");
 
-  strcpy(prio, "NONE:+VERS-TLS-ALL:+SIGN-ALL:+COMP-NULL");
+  strcpy(prio, "NONE:+VERS-TLS-ALL:+SIGN-ALL:+COMP-NULL:+CURVE-ALL");
 
   while (token)
   {
@@ -1062,62 +1043,10 @@ static int ma_gnutls_set_ciphers(gnutls_session_t ssl,
   return gnutls_priority_set_direct(ssl, prio , &err);
 }
 
-static int ma_tls_load_cert(const char *filename,
-                            enum ma_pem_type type,
-                            const char *password,
-                            void *ptr)
-{
-  gnutls_datum_t data;
-  int rc;
-
-  data.data= 0;
-
-  if ((rc= gnutls_load_file(filename, &data)) < 0)
-    goto error;
-
-  switch(type) {
-  case MA_TLS_PEM_CERT:
-  {
-    gnutls_x509_crt_t cert;
-    if ((rc= gnutls_x509_crt_init(&cert)) < 0)
-      goto error;
-    if ((rc= gnutls_x509_crt_import(cert, &data, GNUTLS_X509_FMT_PEM)))
-    {
-      gnutls_x509_crt_deinit(cert);
-      goto error;
-    }
-    *((gnutls_x509_crt_t *)ptr)= cert;
-    return 0;
-  }
-  case MA_TLS_PEM_KEY:
-  {
-    gnutls_x509_privkey_t key;
-    if ((rc= gnutls_x509_privkey_init(&key)) < 0)
-      goto error;
-    if ((rc= gnutls_x509_privkey_import2(key, &data, 
-                                         GNUTLS_X509_FMT_PEM,
-                                         password, 0)) < 0)
-    {
-      gnutls_x509_privkey_deinit(key);
-      goto error;
-    }
-    *((gnutls_x509_privkey_t *)ptr)= key;
-  }
-  default:
-    break;
-  }
-error:
-  if (data.data)
-    gnutls_free(data.data);
-  return rc;
-}
-
 static int ma_tls_set_certs(MYSQL *mysql,
                             gnutls_certificate_credentials_t ctx)
 {
   int  ssl_error= 0;
-  gnutls_x509_privkey_t key= 0;
-  gnutls_x509_crt_t cert= 0;
 
   if (mysql->options.ssl_ca)
   {
@@ -1128,6 +1057,23 @@ static int ma_tls_set_certs(MYSQL *mysql,
     if (ssl_error < 0)
       goto error;
   }
+
+  if (mysql->options.ssl_capath)
+  {
+    ssl_error=  gnutls_certificate_set_x509_trust_dir(ctx,
+                                                      mysql->options.ssl_capath,
+                                                      GNUTLS_X509_FMT_PEM);
+    if (ssl_error < 0)
+      goto error;
+  }
+
+  if (!mysql->options.ssl_ca && !mysql->options.ssl_capath)
+  {
+    ssl_error= gnutls_certificate_set_x509_system_trust(ctx);
+    if (ssl_error < 0)
+      goto error;
+  }
+
   gnutls_certificate_set_verify_function(ctx,
                                          my_verify_callback);
 
@@ -1135,53 +1081,20 @@ static int ma_tls_set_certs(MYSQL *mysql,
   {
     char *keyfile= mysql->options.ssl_key;
     char *certfile= mysql->options.ssl_cert;
-    unsigned char key_id1[65], key_id2[65];
-    size_t key_id1_len, key_id2_len;
-
+    
     if (!certfile)
       certfile= keyfile;
     else if (!keyfile)
       keyfile= certfile;
 
-    if ((ssl_error= ma_tls_load_cert(keyfile, MA_TLS_PEM_KEY,
-                                     mysql->options.extension ?
-                                     mysql->options.extension->tls_pw : NULL,
-                                     &key)) < 0)
-      goto error;
-    if ((ssl_error= ma_tls_load_cert(certfile, MA_TLS_PEM_CERT,
-                                     NULL, &cert)) < 0)
-      goto error;
-
-    /* check if private key corresponds to certificate */
-    key_id1_len= key_id2_len= sizeof(key_id1);
-    if ((ssl_error= gnutls_x509_crt_get_key_id(cert, 0, 
-                                               key_id1, &key_id1_len)) < 0 ||
-        (ssl_error= gnutls_x509_privkey_get_key_id(key, 0, 
-                                                   key_id2, &key_id2_len)) < 0)
-      goto error;
-
-    if (key_id1_len != key_id2_len ||
-        memcmp(key_id1, key_id2, key_id1_len) != 0)
-    {
-      ssl_error= GNUTLS_E_CERTIFICATE_KEY_MISMATCH;
-      goto error;
-    }
-
     /* load cert/key into context */
-    if ((ssl_error= gnutls_certificate_set_x509_key(ctx,
-                                                    &cert,
-                                                    1,
-                                                    key)) < 0)
+    if ((ssl_error= gnutls_certificate_set_x509_key_file2(ctx,
+                                certfile, keyfile, GNUTLS_X509_FMT_PEM,
+                                 mysql->options.extension ? mysql->options.extension->tls_pw : NULL, 0)) < 0)
       goto error;
   }
 
-
-  return ssl_error;
 error:
-  if (key)
-    gnutls_x509_privkey_deinit(key);
-  if (cert)
-    gnutls_x509_crt_deinit(cert);
   return ssl_error;
 }
 
@@ -1190,7 +1103,6 @@ void *ma_tls_init(MYSQL *mysql)
   gnutls_session_t ssl= NULL;
   gnutls_certificate_credentials_t ctx;
   int ssl_error= 0;
-  struct st_gnutls_data *data= NULL;
 
   pthread_mutex_lock(&LOCK_gnutls_config);
 
@@ -1200,14 +1112,10 @@ void *ma_tls_init(MYSQL *mysql)
   if ((ssl_error= ma_tls_set_certs(mysql, ctx)) < 0)
     goto error;
 
-  if ((ssl_error = gnutls_init(&ssl, GNUTLS_CLIENT & GNUTLS_NONBLOCK)) < 0)
+  if ((ssl_error = gnutls_init(&ssl, GNUTLS_CLIENT | GNUTLS_NONBLOCK | GNUTLS_NO_SIGNAL)) < 0)
     goto error;
 
-  if (!(data= (struct st_gnutls_data *)calloc(1, sizeof(struct st_gnutls_data))))
-    goto error;
-
-  data->mysql= mysql;
-  gnutls_session_set_ptr(ssl, (void *)data);
+  gnutls_session_set_ptr(ssl, (void *)mysql);
   /*
   gnutls_certificate_set_retrieve_function2(GNUTLS_xcred, client_cert_callback);
  */
@@ -1223,8 +1131,8 @@ void *ma_tls_init(MYSQL *mysql)
   pthread_mutex_unlock(&LOCK_gnutls_config);
   return (void *)ssl;
 error:
-  free_gnutls_data(data);
   ma_tls_set_error(mysql, ssl, ssl_error);
+  gnutls_certificate_free_credentials(ctx);
   if (ssl)
     gnutls_deinit(ssl);
   pthread_mutex_unlock(&LOCK_gnutls_config);
@@ -1257,12 +1165,9 @@ my_bool ma_tls_connect(MARIADB_TLS *ctls)
 {
   gnutls_session_t ssl = (gnutls_session_t)ctls->ssl;
   my_bool blocking;
-  MYSQL *mysql;
+  MYSQL *mysql= (MYSQL *)gnutls_session_get_ptr(ssl);
   MARIADB_PVIO *pvio;
   int ret;
-  struct st_gnutls_data *data;
-  data= (struct st_gnutls_data *)gnutls_session_get_ptr(ssl);
-  mysql= data->mysql;
 
   if (!mysql)
     return 1;
@@ -1273,6 +1178,7 @@ my_bool ma_tls_connect(MARIADB_TLS *ctls)
   if (!(blocking= pvio->methods->is_blocking(pvio)))
     pvio->methods->blocking(pvio, TRUE, 0);
 
+
 #ifdef GNUTLS_EXTERNAL_TRANSPORT
   /* we don't use GnuTLS read/write functions */
   gnutls_transport_set_ptr(ssl, pvio);
@@ -1290,11 +1196,14 @@ my_bool ma_tls_connect(MARIADB_TLS *ctls)
 
   if (ret < 0)
   {
-    ma_tls_set_error(mysql, ssl, ret);
+    /* If error message was not set while calling certification callback function,
+       use default error message (which is not very descriptive */
+    if (!mysql_errno(mysql))
+      ma_tls_set_error(mysql, ssl, ret);
+
+    ma_tls_close(ctls);
+
     /* restore blocking mode */
-    gnutls_deinit((gnutls_session_t )ctls->ssl);
-    free_gnutls_data(data);
-    ctls->ssl= NULL;
     if (!blocking)
       pvio->methods->blocking(pvio, FALSE, 0);
     return 1;
@@ -1354,12 +1263,40 @@ ssize_t ma_tls_read_async(MARIADB_PVIO *pvio, const uchar *buffer, size_t length
 
 ssize_t ma_tls_read(MARIADB_TLS *ctls, const uchar* buffer, size_t length)
 {
-  return gnutls_record_recv((gnutls_session_t )ctls->ssl, (void *)buffer, length);
+  ssize_t rc;
+  MARIADB_PVIO *pvio= ctls->pvio;
+
+  while ((rc= gnutls_record_recv((gnutls_session_t)ctls->ssl, (void *)buffer, length)) <= 0)
+  {
+    if (rc != GNUTLS_E_AGAIN && rc != GNUTLS_E_INTERRUPTED)
+      break;
+    if (pvio->methods->wait_io_or_timeout(pvio, TRUE, pvio->mysql->options.read_timeout) < 1)
+      break;
+  }
+  if (rc <= 0) {
+    MYSQL *mysql= (MYSQL *)gnutls_session_get_ptr(ctls->ssl);
+    ma_tls_set_error(mysql, ctls->ssl, rc);
+  }
+  return rc;
 }
 
 ssize_t ma_tls_write(MARIADB_TLS *ctls, const uchar* buffer, size_t length)
 { 
-  return gnutls_record_send((gnutls_session_t )ctls->ssl, (void *)buffer, length);
+  ssize_t rc;
+  MARIADB_PVIO *pvio= ctls->pvio;
+
+  while ((rc= gnutls_record_send((gnutls_session_t)ctls->ssl, (void *)buffer, length)) <= 0)
+  {
+    if (rc != GNUTLS_E_AGAIN && rc != GNUTLS_E_INTERRUPTED)
+      break;
+    if (pvio->methods->wait_io_or_timeout(pvio, TRUE, pvio->mysql->options.write_timeout) < 1)
+      break;
+  }
+  if (rc <= 0) {
+    MYSQL *mysql= (MYSQL *)gnutls_session_get_ptr(ctls->ssl);
+    ma_tls_set_error(mysql, ctls->ssl, rc);
+  }
+  return rc;
 }
 
 my_bool ma_tls_close(MARIADB_TLS *ctls)
@@ -1367,14 +1304,11 @@ my_bool ma_tls_close(MARIADB_TLS *ctls)
   if (ctls->ssl)
   {
     gnutls_certificate_credentials_t ctx;
-    struct st_gnutls_data *data=
-      (struct st_gnutls_data *)gnutls_session_get_ptr(ctls->ssl);
     /* this would be the correct way, however can't detect afterwards
        if the socket is closed or not, so we don't send encrypted 
        finish alert.
     rc= gnutls_bye((gnutls_session_t )ctls->ssl, GNUTLS_SHUT_WR);
     */
-    free_gnutls_data(data);
     gnutls_credentials_get(ctls->ssl, GNUTLS_CRD_CERTIFICATE, (void **)&ctx);
     gnutls_certificate_free_keys(ctx);
     gnutls_certificate_free_cas(ctx);
@@ -1389,7 +1323,7 @@ my_bool ma_tls_close(MARIADB_TLS *ctls)
 
 int ma_tls_verify_server_cert(MARIADB_TLS *ctls __attribute__((unused)))
 {
-  /* server verification is already handled before */
+  /* server verification is already handled before during handshake */
   return 0;
 }
 
@@ -1411,77 +1345,41 @@ const char *ma_tls_get_cipher(MARIADB_TLS *ctls)
 static int my_verify_callback(gnutls_session_t ssl)
 {
   unsigned int status= 0;
-  const gnutls_datum_t *cert_list;
-  unsigned int cert_list_size;
-  struct st_gnutls_data *data= (struct st_gnutls_data *)gnutls_session_get_ptr(ssl);
-  MYSQL *mysql;
-  MARIADB_PVIO *pvio;
-  
-  gnutls_x509_crt_t cert;
-  const char *hostname;
-
-  mysql= data->mysql;
-  pvio= mysql->net.pvio;
-
-  /* read hostname */
-  hostname = mysql->host;
-
-  /* This verification function uses the trusted CAs in the credentials
-   * structure. So you must have installed one or more CA certificates.
-   */
-  if ((mysql->client_flag & CLIENT_SSL_VERIFY_SERVER_CERT)  &&
-        gnutls_certificate_verify_peers2 (ssl, &status) < 0)
-  {
-    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, "CA verification failed");
-    return GNUTLS_E_CERTIFICATE_ERROR;
-  }
-
-  if (status & GNUTLS_CERT_INVALID)
-  {
-    char errbuf[100];
-    snprintf(errbuf, 99, "CA Verification failed (Status: %d)", status);
-    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, errbuf);
-    return GNUTLS_E_CERTIFICATE_ERROR;
-  }
-  /* Up to here the process is the same for X.509 certificates and
-   * OpenPGP keys. From now on X.509 certificates are assumed. This can
-   * be easily extended to work with openpgp keys as well.
-   */
-  if (gnutls_certificate_type_get (ssl) != GNUTLS_CRT_X509)
-  {
-    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, "Expected X509 certificate");
-    return GNUTLS_E_CERTIFICATE_ERROR;
-  }
-  if (gnutls_x509_crt_init (&cert) < 0)
-  {
-    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, "Error during certificate initialization");
-    return GNUTLS_E_CERTIFICATE_ERROR;
-  }
-  cert_list = gnutls_certificate_get_peers (ssl, &cert_list_size);
-  if (cert_list == NULL)
-  {
-    gnutls_x509_crt_deinit (cert);
-    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, "No certificate found");
-    return GNUTLS_E_CERTIFICATE_ERROR;
-  }
-  if (gnutls_x509_crt_import (cert, &cert_list[0], GNUTLS_X509_FMT_DER) < 0)
-  {
-    gnutls_x509_crt_deinit (cert);
-    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, "Unknown SSL error");
-    return GNUTLS_E_CERTIFICATE_ERROR;
-  }
-
-  if ((mysql->client_flag & CLIENT_SSL_VERIFY_SERVER_CERT) &&
-      !gnutls_x509_crt_check_hostname (cert, hostname))
-  {
-    gnutls_x509_crt_deinit (cert);
-    pvio->set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN, "Hostname in certificate doesn't match");
-    return GNUTLS_E_CERTIFICATE_ERROR;
-  }
-  gnutls_x509_crt_deinit (cert);
-  /* notify gnutls to continue handshake normally */
+  MYSQL *mysql= (MYSQL *)gnutls_session_get_ptr(ssl);
 
   CLEAR_CLIENT_ERROR(mysql);
+
+  if ((mysql->options.extension->tls_verify_server_cert))
+  {
+    const char *hostname= mysql->host;
+
+    if (gnutls_certificate_verify_peers3 (ssl, hostname, &status) < 0)
+      return GNUTLS_E_CERTIFICATE_ERROR;
+  } else {
+    if (gnutls_certificate_verify_peers2 (ssl, &status) < 0)
+      return GNUTLS_E_CERTIFICATE_ERROR;
+  }
+  if (status & GNUTLS_CERT_INVALID)
+  {
+    gnutls_datum_t out;
+    int type;
+    /* accept self signed certificates if we don't have to verify server cert */
+    if (!(mysql->options.extension->tls_verify_server_cert) &&
+        (status & GNUTLS_CERT_SIGNER_NOT_FOUND))
+      return 0;
+
+    /* gnutls default error message "certificate validation failed" isn't very
+       descriptive, so we provide more information about the error here */
+    type= gnutls_certificate_type_get(ssl);
+    gnutls_certificate_verification_status_print(status, type, &out, 0);
+    my_set_error(mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN,
+                 ER(CR_SSL_CONNECTION_ERROR), out.data);
+    gnutls_free(out.data);
+    
+    return GNUTLS_E_CERTIFICATE_ERROR;
+  }
+  
+
   return 0;
 }
 
@@ -1491,13 +1389,11 @@ unsigned int ma_tls_get_finger_print(MARIADB_TLS *ctls, char *fp, unsigned int l
   size_t fp_len= len;
   const gnutls_datum_t *cert_list;
   unsigned int cert_list_size;
-  struct st_gnutls_data *data;
 
   if (!ctls || !ctls->ssl)
     return 0;
 
-  data= (struct st_gnutls_data *)gnutls_session_get_ptr(ctls->ssl);
-  mysql= (MYSQL *)data->mysql;
+  mysql= (MYSQL *)gnutls_session_get_ptr(ctls->ssl);
 
   cert_list = gnutls_certificate_get_peers (ctls->ssl, &cert_list_size);
   if (cert_list == NULL)
@@ -1526,4 +1422,9 @@ int ma_tls_get_protocol_version(MARIADB_TLS *ctls)
 
   return gnutls_protocol_get_version(ctls->ssl) - 1;
 }
+
+void ma_tls_set_connection(MYSQL *mysql)
+{
+  (void)gnutls_session_set_ptr(mysql->net.pvio->ctls->ssl, (void *)mysql);
+}
 #endif /* HAVE_GNUTLS */
